@@ -2,24 +2,17 @@ __version__ = "0.1.2"
 
 
 from jax_metrics.losses import Loss
-from jax_metrics.metrics import (
-    AuxLosses,
-    AuxMetrics,
-    Losses,
-    LossesAndMetrics,
-    Metric,
-    Metrics,
-)
+from jax_metrics.metrics import AuxMetrics, Losses, Metric, Metrics, SumMetric
 from jax_metrics.types import Named
 
 from . import losses, metrics, regularizers
 
 __all__ = [
     "Loss",
-    "LossesAndMetrics",
     "Losses",
     "Metrics",
     "Metric",
+    "SumMetric",
     "Named",
     "losses",
     "metrics",
